@@ -1,3 +1,5 @@
+{-# LANGUAGE NoImplicitPrelude #-}
+
 {-
    **************************************************************
     BNF Converter Module
@@ -22,6 +24,8 @@
 
 module BNFC.Backend.CPP.PrettyPrinter (cf2CPPPrinter, prRender) where
 
+import Prelude'
+
 import BNFC.CF
 import BNFC.Utils ((+++))
 import BNFC.Backend.Common
@@ -40,8 +44,9 @@ cf2CPPPrinter useStl inPackage cf =
 
 positionRules :: CF -> [(Cat,[Rule])]
 positionRules cf =
-      [(cat,[Rule (show cat) cat [Left catString, Left catInteger]]) |
-        cat <- filter (isPositionCat cf) $ fst (unzip (tokenPragmas cf))]
+  [ (TokenCat cat, [Rule cat (TokenCat cat) $ map (Left . TokenCat) [catString, catInteger]])
+  | cat <- filter (isPositionCat cf) $ map fst (tokenPragmas cf)
+  ]
 
 {- **** Header (.H) File Methods **** -}
 
@@ -82,14 +87,14 @@ mkHFile useStl inPackage cf groups = unlines
     "  /* The following are simple heuristics for rendering terminals */",
     "  /* You may wish to change them */",
     "  void render(Char c);",
-    "  void render(String s);",
-    if useStl then "void render(char* s);" else "",
+    if useStl then "  void render(String s);" else "",
+    "  void render(const char *s);",
     "  void indent(void);",
     "  void backup(void);",
     " public:",
     "  PrintAbsyn(void);",
     "  ~PrintAbsyn(void);",
-    "  char* print(Visitable* v);"
+    "  char *print(Visitable *v);"
    ]
   hdef = nsDefine inPackage "PRINTER_HEADER"
   classFooter = unlines $
@@ -102,14 +107,17 @@ mkHFile useStl inPackage cf groups = unlines
    ] ++ ["  void visit" ++ t ++ "(String s);" | t <- tokenNames cf] ++
    [
     " protected:",
-    "  void inline bufAppend(const char* s)",
+    "  char *buf_;",
+    "  int cur_, buf_size;",
+    "",
+    "  void inline bufAppend(const char *s)",
     "  {",
     "    int len = strlen(s);",
-    "    while (cur_ + len > buf_size)",
+    "    while (cur_ + len >= buf_size)",
     "    {",
     "      buf_size *= 2; /* Double the buffer size */",
-    "      resizeBuffer();",
     "    }",
+    "    resizeBuffer();",
     "    for(int n = 0; n < len; n++)",
     "    {",
     "      buf_[cur_ + n] = s[n];",
@@ -117,9 +125,10 @@ mkHFile useStl inPackage cf groups = unlines
     "    cur_ += len;",
     "    buf_[cur_] = 0;",
     "  }",
+    "",
     "  void inline bufAppend(const char c)",
     "  {",
-    "    if (cur_ == buf_size)",
+    "    if (cur_ >= buf_size)",
     "    {",
     "      buf_size *= 2; /* Double the buffer size */",
     "      resizeBuffer();",
@@ -128,17 +137,24 @@ mkHFile useStl inPackage cf groups = unlines
     "    cur_++;",
     "    buf_[cur_] = 0;",
     "  }",
+    "",
     if useStl then render (nest 2 bufAppendString) else "",
     "  void inline bufReset(void)",
     "  {",
-    "    cur_ = 0;",
+    "    if (buf_) free(buf_);",
     "    buf_size = " ++ nsDefine inPackage "BUFFER_INITIAL" ++ ";",
-    "    resizeBuffer();",
+    "    buf_ = (char *) malloc(buf_size);",
+    "    if (!buf_) {",
+    "      fprintf(stderr, \"Error: Out of memory while allocating buffer!\\n\");",
+    "      exit(1);",
+    "    }",
     "    memset(buf_, 0, buf_size);",
+    "    cur_ = 0;",
     "  }",
+    "",
     "  void inline resizeBuffer(void)",
     "  {",
-    "    char* temp = (char*) malloc(buf_size);",
+    "    char *temp = (char *) malloc(buf_size);",
     "    if (!temp)",
     "    {",
     "      fprintf(stderr, \"Error: Out of memory while attempting to grow buffer!\\n\");",
@@ -151,8 +167,6 @@ mkHFile useStl inPackage cf groups = unlines
     "    }",
     "    buf_ = temp;",
     "  }",
-    "  char *buf_;",
-    "  int cur_, buf_size;",
     "};",
     ""
    ]
@@ -160,7 +174,7 @@ mkHFile useStl inPackage cf groups = unlines
   bufAppendString =
       "void inline bufAppend(String str)"
       $$ codeblock 2
-          [ "const char* s = str.c_str();"
+          [ "const char *s = str.c_str();"
           , "bufAppend(s);"]
   showHeader = unlines
    [
@@ -170,7 +184,7 @@ mkHFile useStl inPackage cf groups = unlines
     " public:",
     "  ShowAbsyn(void);",
     "  ~ShowAbsyn(void);",
-    "  char* show(Visitable* v);"
+    "  char *show(Visitable *v);"
    ]
   footer = unlines
    [
@@ -183,7 +197,7 @@ mkHFile useStl inPackage cf groups = unlines
 prDataH :: (Cat, [Rule]) -> String
 prDataH (cat, rules) =
  if isList cat
- then concat ["  void visit", cl, "(", cl, "* p);\n"]
+ then concat ["  void visit", cl, "(", cl, " *p);\n"]
  else abstract ++ concatMap prRuleH rules
  where
    cl = identCat (normCat cat)
@@ -239,13 +253,14 @@ mkCFile useStl inPackage cf groups = concat
       "{",
       "}",
       "",
-      "char* PrintAbsyn::print(Visitable *v)",
+      "char *PrintAbsyn::print(Visitable *v)",
       "{",
       "  _i_ = 0; _n_ = 0;",
       "  bufReset();",
       "  v->accept(this);",
       "  return buf_;",
-      "}"
+      "}",
+      ""
      ]
     showEntries = unlines
      [
@@ -259,12 +274,13 @@ mkCFile useStl inPackage cf groups = concat
       "{",
       "}",
       "",
-      "char* ShowAbsyn::show(Visitable *v)",
+      "char *ShowAbsyn::show(Visitable *v)",
       "{",
       "  bufReset();",
       "  v->accept(this);",
       "  return buf_;",
-      "}"
+      "}",
+      ""
      ]
     printBasics = unlines
      [
@@ -274,24 +290,28 @@ mkCFile useStl inPackage cf groups = concat
       "  sprintf(tmp, \"%d\", i);",
       "  bufAppend(tmp);",
       "}",
+      "",
       "void PrintAbsyn::visitDouble(Double d)",
       "{",
       "  char tmp[16];",
       "  sprintf(tmp, \"%g\", d);",
       "  bufAppend(tmp);",
       "}",
+      "",
       "void PrintAbsyn::visitChar(Char c)",
       "{",
       "  bufAppend('\\'');",
       "  bufAppend(c);",
       "  bufAppend('\\'');",
       "}",
+      "",
       "void PrintAbsyn::visitString(String s)",
       "{",
       "  bufAppend('\\\"');",
       "  bufAppend(s);",
       "  bufAppend('\\\"');",
       "}",
+      "",
       "void PrintAbsyn::visitIdent(String s)",
       "{",
       "  render(s);",
@@ -359,26 +379,28 @@ mkCFile useStl inPackage cf groups = concat
 
 {- **** Pretty Printer Methods **** -}
 
---Generates methods for the Pretty Printer
+-- | Generates methods for the Pretty Printer.
 prPrintData :: Bool -> Maybe String -> CF -> (Cat, [Rule]) -> String
 prPrintData True {- use STL -} _ _ (cat@(ListCat _), rules) =
     render $ genPrintVisitorList (cat, rules)
 prPrintData False {- use STL -} _ _ (cat@(ListCat _), rules) =
     genPrintVisitorListNoStl (cat, rules)
-prPrintData _ inPackage cf (cat, rules) = -- Not a list
- -- a position token
- if isPositionCat cf cat then unlines [
-   "void PrintAbsyn::visit" ++ show cat ++ "(" ++ show cat ++ "* p)",
-   "{",
-   "  visitIdent(p->string_);",
-   "}"
-   ]
- else abstract ++ concatMap (prPrintRule inPackage) rules
- where
-   cl = identCat (normCat cat)
-   abstract = case lookupRule (show cat) rules of
+-- Not a list :
+prPrintData _ _inPackage cf (TokenCat cat, _rules) | isPositionCat cf cat = unlines $
+  -- a position token
+  [ "void PrintAbsyn::visit" ++ show cat ++ "(" ++ show cat ++ " *p)"
+  , "{"
+  , "  visitIdent(p->string_);"
+  , "}"
+  , ""
+  ]
+prPrintData _ inPackage _cf (cat, rules) = -- Not a list
+    abstract ++ concatMap (prPrintRule inPackage) rules
+  where
+  cl = identCat (normCat cat)
+  abstract = case lookupRule (show cat) rules of
     Just _ -> ""
-    Nothing ->  "void PrintAbsyn::visit" ++ cl ++ "(" ++ cl ++ "*p) {} //abstract class\n\n"
+    Nothing ->  "void PrintAbsyn::visit" ++ cl ++ "(" ++ cl +++ "*p) {} //abstract class\n\n"
 
 -- | Generate pretty printer visitor for a list category:
 --
@@ -437,7 +459,7 @@ genPrintVisitorListNoStl :: (Cat, [Rule]) -> String
 genPrintVisitorListNoStl (cat@(ListCat c), rules) = unlines
     [ "void PrintAbsyn::visit" ++ cl ++ "("++ cl ++ " *" ++ vname ++ ")"
     , "{"
-    , "  while(" ++ vname ++ "!= 0)"
+    , "  while(" ++ vname +++ "!= 0)"
     , "  {"
     , "    if (" ++ vname ++ "->" ++ vname ++ "_ == 0)"
     , "    {"
@@ -474,14 +496,15 @@ genPrintVisitorListNoStl _ = error "genPrintVisitorListNoStl expects a ListCat"
 prPrintRule :: Maybe String -> Rule -> String
 prPrintRule inPackage r@(Rule fun _ cats) | isProperLabel fun = unlines
   [
-   "void PrintAbsyn::visit" ++ fun ++ "(" ++ fun ++ "*" +++ fnm ++ ")",
+   "void PrintAbsyn::visit" ++ fun ++ "(" ++ fun +++ "*" ++ fnm ++ ")",
    "{",
    "  int oldi = _i_;",
    lparen,
    cats',
    rparen,
    "  _i_ = oldi;",
-   "}\n"
+   "}",
+   ""
   ]
    where
     p = precRule r
@@ -498,7 +521,7 @@ prPrintCat _ (Right t) = "  render(" ++ t' ++ ");\n"
   where t' = snd (renderCharOrString t)
 prPrintCat fnm (Left (c, nt))
   | isTokenCat c  = "  visit" ++ funName c ++ "(" ++ fnm ++ "->" ++ render nt ++ ");\n"
-  | isList c            = "  if(" ++ fnm ++ "->" ++ render nt ++ ") {" ++ accept ++ "}"
+  | isList c            = "  if(" ++ fnm ++ "->" ++ render nt ++ ") {" ++ accept ++ "}\n"
   | otherwise           = "  " ++ accept ++ "\n"
   where
     accept
@@ -563,13 +586,13 @@ prShowData _ (cat, rules) =  --Not a list:
     cl = identCat (normCat cat)
     abstract = case lookupRule (show cat) rules of
       Just _ -> ""
-      Nothing ->  "void ShowAbsyn::visit" ++ cl ++ "(" ++ cl ++ "* p) {} //abstract class\n\n"
+      Nothing ->  "void ShowAbsyn::visit" ++ cl ++ "(" ++ cl ++ " *p) {} //abstract class\n\n"
 
 --This prints all the methods for Abstract Syntax tree rules.
 prShowRule :: Rule -> String
 prShowRule (Rule fun _ cats) | isProperLabel fun = concat
   [
-   "void ShowAbsyn::visit" ++ fun ++ "(" ++ fun ++ "*" +++ fnm ++ ")\n",
+   "void ShowAbsyn::visit" ++ fun ++ "(" ++ fun +++ "*" ++ fnm ++ ")\n",
    "{\n",
    lparen,
    "  bufAppend(\"" ++ fun ++ "\");\n",
@@ -692,6 +715,7 @@ prRender useStl = unlines
       "     bufAppend(' ');",
       "  }",
       "}",
+      "",
       let renderString = "void PrintAbsyn::render(String s_)" $$ codeblock 2
                              [ "const char *s = s_.c_str() ;"
                              , "if(strlen(s) > 0)"
@@ -699,7 +723,7 @@ prRender useStl = unlines
                                  [ "bufAppend(s);"
                                  , "bufAppend(' ');" ] ]
       in if useStl then render renderString else "",
-      "void PrintAbsyn::render(char* s)",
+      "void PrintAbsyn::render(const char *s)",
       "{",
       "  if(strlen(s) > 0)",
       "  {",
@@ -707,6 +731,7 @@ prRender useStl = unlines
       "    bufAppend(' ');",
       "  }",
       "}",
+      "",
       "void PrintAbsyn::indent()",
       "{",
       "  int n = _n_;",
@@ -716,6 +741,7 @@ prRender useStl = unlines
       "    n--;",
       "  }",
       "}",
+      "",
       "void PrintAbsyn::backup()",
       "{",
       "  if (buf_[cur_ - 1] == ' ')",
@@ -723,5 +749,6 @@ prRender useStl = unlines
       "    buf_[cur_ - 1] = 0;",
       "    cur_--;",
       "  }",
-      "}"
+      "}",
+      ""
   ]
