@@ -1,7 +1,11 @@
-{-# LANGUAGE PatternGuards, DeriveFunctor #-}
+{-# LANGUAGE DeriveFunctor #-}
+{-# LANGUAGE LambdaCase #-}
+{-# LANGUAGE PatternGuards #-}
+{-# LANGUAGE RecordWildCards #-}
+
 {-
     BNF Converter: Abstract syntax
-    Copyright (C) 2004  Author:  Markus Forberg, Michael Pellauer, Aarne Ranta
+    Copyright (C) 2004  Author:  Markus Forsberg, Michael Pellauer, Aarne Ranta
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -15,20 +19,21 @@
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1335, USA
 -}
 
 module BNFC.CF (
             -- Types.
             CF,
-            CFG(..), pragmasOfCF, -- ...
+            CFG(..),
             Rule, Rul(..), lookupRule,
             Pragma(..),
             Exp(..),
             Literal,
             Symbol,
             KeyWord,
-            Cat(..), strToCat,
+            Cat(..), strToCat, catToStr,
+            TokenCat,
             catString, catInteger, catDouble, catChar, catIdent,
             NonTerminal,
             Fun,
@@ -41,6 +46,7 @@ module BNFC.CF (
             -- Literal categories, constants,
             firstCat,       -- the first value category in the grammar.
             firstEntry,     -- the first entry or the first value category
+            baseTokenCatNames,  -- "Char", "Double", "Integer", "String"
             specialCats,    -- ident
             specialCatsP,   -- all literals
             specialData,    -- special data
@@ -53,22 +59,19 @@ module BNFC.CF (
             allEntryPoints,
             reservedWords,
             cfTokens,
-            symbols,
             literals,
-            reversibleCats,
             findAllReversibleCats, -- find all reversible categories
             identCat,       -- transforms '[C]' to ListC (others, unchanged).
             isParsable,
-            rulesOfCF,      -- All rules of a grammar.
             rulesForCat,    -- rules for a given category
             rulesForNormalizedCat,    -- rules for a given category
             ruleGroups,     -- Categories are grouped with their rules.
             ruleGroupsInternals, --As above, but includes internal cats.
-            notUniqueNames, -- list of not unique names (replaces the following 2)
---          notUniqueFuns,   -- Returns a list of function labels that are not unique.
---            badInheritence, -- Returns a list of all function labels that can cause problems in languages with inheritence.
+            allNames,        -- Checking for non-unique names, like @Name. Name ::= Ident;@.
+            filterNonUnique,
             isList,         -- Checks if a category is a list category.
             isTokenCat,
+            sameCat,
             -- Information functions for list functions.
             isNilFun,       -- empty list function? ([])
             isOneFun,       -- one element list function? (:[])
@@ -105,7 +108,6 @@ module BNFC.CF (
             cf2cfp,
             cfp2cf,
             trivialProf,
-            rulesOfCFP,
             funRuleP, ruleGroupsP, allCatsP, allEntryPointsP
            ) where
 
@@ -120,49 +122,56 @@ import qualified AbsBNF
 import ErrM
 
 -- | A context free grammar consists of a set of rules and some extended
--- information (e.g. pragmas, literals, symbols, keywords)
+-- information (e.g. pragmas, literals, symbols, keywords).
+
 type CF = CFG Fun
 
 -- | A rule consists of a function name, a main category and a sequence of
 -- terminals and non-terminals.
--- function_name . Main_Cat ::= sequence
+-- @
+--   function_name . Main_Cat ::= sequence
+-- @
+
 type Rule = Rul Fun
 
--- | Polymorphic rule type for common type signatures for CF and CFP
-data Rul function = Rule { funRule :: function
-                           -- ^ The function (semantic action) of a
-                           -- rule. In order to be able to generate
-                           -- data types this must be a constructor
-                           -- (or an identity function).
-                         , valCat :: Cat -- ^ The value category
-                         , rhsRule :: [Either Cat String]
-                           -- ^ The list of Terminals/NonTerminals in
-                           -- the right-hand-side of a rule.
-                         }
-                deriving (Eq,Functor)
+-- | Polymorphic rule type for common type signatures for CF and CFP.
+
+data Rul function = Rule
+  { funRule :: function
+      -- ^ The function (semantic action) of a rule.
+      --   In order to be able to generate data types this must be a constructor
+      --   (or an identity function).
+  , valCat  :: Cat
+      -- ^ The value category, i.e., the defined non-terminal.
+  , rhsRule :: [Either Cat String]
+      -- ^ The sentential form, i.e.,
+      --   the list of (non)terminals in the right-hand-side of a rule.
+  } deriving (Eq, Functor)
 
 instance (Show function) => Show (Rul function) where
   show (Rule f cat rhs) =
       unwords (show f : "." : show cat : "::=" : map (either show id) rhs)
 
--- | Polymorphic CFG type for common type signatures for CF and CFP
-newtype CFG function = CFG { unCFG :: (Exts,[Rul function]) }
-                deriving (Functor)
+-- | Polymorphic CFG type for common type signatures for CF and CFP.
+
+data CFG function = CFG
+    { cfgPragmas        :: [Pragma]
+    , cfgLiterals       :: [Literal]  -- ^ @Char, String, Ident, Integer, Double@.
+                                      --   @String@s are quoted strings,
+                                      --   and @Ident@s are unquoted.
+    , cfgSymbols        :: [Symbol]   -- ^ Symbols in the grammar, e.g. “*”, '->'.
+    , cfgKeywords       :: [KeyWord]  -- ^ Reserved words, e.g. 'if' 'while'.
+    , cfgReversibleCats :: [Cat]      -- ^ Categories that can be made left-recursive.
+    , cfgRules          :: [Rul function]
+    } deriving (Functor)
 
 
 instance (Show function) => Show (CFG function) where
-  show (CFG (_,rules)) = unlines $ map show rules
+  show CFG{..} = unlines $ map show cfgRules
 
-type Exts = ([Pragma],Info)
 
--- | Info is information extracted from the CF, for easy access.
--- Literals - Char, String, Ident, Integer, Double
---            Strings are quoted strings, and Ident are unquoted.
--- Symbols  - symbols in the grammar, e.g. “*”, '->'.
--- KeyWord  - reserved words, e.g. 'if' 'while'
-type Info = ([Literal],[Symbol],[KeyWord],[Cat])
+-- | Expressions for function definitions.
 
--- Expressions for function definitions
 data Exp = App String [Exp]
          | LitInt Integer
          | LitDouble Double
@@ -198,7 +207,8 @@ instance Show Exp where
                 | Right es <- listView e2   = Right $ e1:es
             listView x = Left x
 
--- | Pragmas
+-- | Pragmas.
+
 data Pragma = CommentS  String -- ^ for single line comments
             | CommentM (String,String) -- ^  for multiple-line comments.
             | TokenReg String Bool Reg -- ^ for tokens
@@ -211,15 +221,15 @@ data Pragma = CommentS  String -- ^ for single line comments
               deriving (Show)
 
 -- | User-defined regular expression tokens
-tokenPragmas :: CFG f -> [(Cat,Reg)]
-tokenPragmas cf = [(TokenCat name,e) | TokenReg name _ e <- pragmasOfCF cf]
+tokenPragmas :: CFG f -> [(TokenCat,Reg)]
+tokenPragmas cf = [ (name, e) | TokenReg name _ e <- cfgPragmas cf ]
 
--- | The names of all user-defined tokens
+-- | The names of all user-defined tokens.
 tokenNames :: CFG f -> [String]
-tokenNames cf = map (show.fst) (tokenPragmas cf)
+tokenNames cf = map fst (tokenPragmas cf)
 
 layoutPragmas :: CF -> (Bool,[String],[String])
-layoutPragmas cf = let ps = pragmasOfCF cf in (
+layoutPragmas cf = let ps = cfgPragmas cf in (
   not (null [() | LayoutTop  <- ps]),   -- if there's layout betw top-level
   concat [ss | Layout ss     <- ps],    -- layout-block starting words
   concat [ss | LayoutStop ss <- ps]     -- layout-block ending words
@@ -229,7 +239,7 @@ hasLayout :: CF -> Bool
 hasLayout cf = case layoutPragmas cf of
   (t,ws,_) -> t || not (null ws)   -- (True,[],_) means: top-level layout only
 
--- | Literal: Char, String, Ident, Integer, Double
+-- | Literal: builtin-token types Char, String, Ident, Integer, Double.
 type Literal = String
 type Symbol  = String
 type KeyWord = String
@@ -238,50 +248,38 @@ type KeyWord = String
 -- Categories
 ------------------------------------------------------------------------------
 
--- | Categories are the Non-terminals of the grammar.
-data Cat = InternalCat       -- | Internal category, inserted in 1st
-                             -- position in "internal" rules,
-                             -- essentially ensuring that they are
-                             -- never parsed.
-         | Cat String
-         | TokenCat String   -- ^ Token types (like Ident)
-         | ListCat Cat
-         | CoercCat String Integer
+-- | Categories are the non-terminals of the grammar.
+data Cat
+  = InternalCat       -- ^ Internal category, inserted in 1st
+                      --   position in "internal" rules,
+                      --   essentially ensuring that they are never parsed.
+  | Cat String               -- ^ Ordinary non-terminal.
+  | TokenCat TokenCat        -- ^ Token types (like @Ident@, @Integer@, ..., user-defined).
+  | ListCat Cat              -- ^ List non-terminals, e.g., @[Ident]@, @[Exp]@, @[Exp1]@.
+  | CoercCat String Integer  -- ^ E.g. @Exp1@, @Exp2.
   deriving (Eq, Ord)
 
+type TokenCat = String
+
 -- An alias for Cat used in many backends:
 type NonTerminal = Cat
 
 -- | Render category symbols as strings
--- >>> catToStr (Cat "Def")
--- "Def"
--- >>> catToStr (ListCat (Cat "Thing"))
--- "[Thing]"
--- >>> catToStr (CoercCat "Expr" 3)
--- "Expr3"
--- >>> catToStr (ListCat (CoercCat "Expr" 2))
--- "[Expr2]"
--- >>> catToStr (TokenCat "Abc")
--- "Abc"
-catToStr InternalCat = "#"
-catToStr (Cat s) = s
-catToStr (TokenCat s) = s
-catToStr (ListCat c) = "[" ++ show c ++ "]"
-catToStr (CoercCat s i) = s ++ show i
+catToStr :: Cat -> String
+catToStr = \case
+  InternalCat  -> "#"
+  Cat s        -> s
+  TokenCat s   -> s
+  ListCat c    -> "[" ++ catToStr c ++ "]"
+  CoercCat s i -> s ++ show i
 
 instance Show Cat where
   show = catToStr
 
--- | Reads a string into a category. This should only needs to handle
+-- | Reads a string into a category. This should only need to handle
 -- the case of simple categories (with or without coercion) since list
 -- categories are parsed in the grammar already. To be on the safe side here,
 -- we still call the parser function that parses categries.
--- >>> strToCat "Abc" == Cat "Abc"
--- True
--- >>> strToCat "Abc123" == CoercCat "Abc" 123
--- True
--- >>> strToCat "[Expr2]" == ListCat (CoercCat "Expr" 2)
--- True
 strToCat :: String -> Cat
 strToCat "#" = InternalCat
 strToCat s =
@@ -297,16 +295,20 @@ strToCat s =
         cat2cat (AbsBNF.ListCat c) = ListCat (cat2cat c)
 
 -- Build-in categories contants
-catString, catInteger, catDouble, catChar, catIdent :: Cat
-catString  = TokenCat "String"
-catInteger = TokenCat "Integer"
-catDouble  = TokenCat "Double"
-catChar    = TokenCat "Char"
-catIdent   = TokenCat "Ident"
+catString, catInteger, catDouble, catChar, catIdent :: TokenCat
+catString  = "String"
+catInteger = "Integer"
+catDouble  = "Double"
+catChar    = "Char"
+catIdent   = "Ident"
+
+-- | Token categories corresponding to base types.
+baseTokenCatNames :: [TokenCat]
+baseTokenCatNames = [ catChar, catDouble, catInteger, catString ]
 
 -- the parser needs these
-specialCatsP :: [String]
-specialCatsP = words "Ident Integer String Char Double"
+specialCatsP :: [TokenCat]
+specialCatsP = catIdent : baseTokenCatNames
 
 -- | Does the category correspond to a data type?
 isDataCat :: Cat -> Bool
@@ -318,27 +320,20 @@ isDataOrListCat (Cat ('@':_))   = False
 isDataOrListCat (ListCat c)     = isDataOrListCat c
 isDataOrListCat _               = True
 
--- | Categories C1, C2,... (one digit in end) are variants of C. This function
+-- | Categories C1, C2,... (one digit at the end) are variants of C. This function
 -- returns true if two category are variants of the same abstract category.
 -- E.g.
 --
--- >>> eqCat (Cat "Abc") (Cat "Abc")
+-- >>> sameCat (Cat "Abc") (CoercCat "Abc" 44)
 -- True
--- >>> eqCat (CoercCat "Abc" 3) (CoercCat "Abc" 5)
--- True
--- >>> eqCat (CoercCat "Acb" 4) (CoercCat "Abc" 4)
--- False
--- >>> eqCat (Cat "Abc") (CoercCat "Abc" 44)
--- True
-eqCat :: Cat -> Cat -> Bool
-eqCat (CoercCat c1 _) (CoercCat c2 _) = c1 == c2
-eqCat (Cat c1 ) (CoercCat c2 _) = c1 == c2
-eqCat (CoercCat c1 _) (Cat c2) = c1 == c2
-eqCat c1 c2 = c1 == c2
 
+sameCat :: Cat -> Cat -> Bool
+sameCat (CoercCat c1 _) (CoercCat c2 _) = c1 == c2
+sameCat (Cat c1)        (CoercCat c2 _) = c1 == c2
+sameCat (CoercCat c1 _) (Cat c2)        = c1 == c2
+sameCat c1              c2              = c1 == c2
 
-
--- | Removes precendence information. C1 => C, [C2] => [C]
+-- | Removes precedence information. C1 => C, [C2] => [C]
 normCat :: Cat -> Cat
 normCat (ListCat c) = ListCat (normCat c)
 normCat (CoercCat c _) = Cat c
@@ -350,10 +345,6 @@ normCatOfList = normCat . catOfList
 -- | When given a list Cat, i.e. '[C]', it removes the square
 -- brackets, and adds the prefix List, i.e. 'ListC'.  (for Happy and
 -- Latex)
--- >>> identCat (ListCat (Cat "C")) -- [C]
--- "ListC"
--- >>> identCat (CoercCat "C" 3)
--- "C3"
 identCat :: Cat -> String
 identCat (ListCat c) = "List" ++ identCat c
 identCat c = show c
@@ -366,24 +357,20 @@ isTokenCat :: Cat -> Bool
 isTokenCat (TokenCat _) = True
 isTokenCat _            = False
 
--- | Unwraps the list constructor from the category name. Eg. [C1] => C1
--- E.g.
--- >>> catOfList (ListCat (Cat "A"))
--- A
--- >>> catOfList (Cat "B")
--- B
+-- | Unwraps the list constructor from the category name.
+--   E.g. @[C1] => C1@.
 catOfList :: Cat -> Cat
 catOfList (ListCat c) = c
 catOfList c = c
+
 ------------------------------------------------------------------------------
 -- Functions
 ------------------------------------------------------------------------------
+
 -- | Fun is the function name of a rule.
 type Fun     = String
--- | Either Cat or Fun
 
-
--- | Is this function just a coercion? (Ie. the identity)
+-- | Is this function just a coercion? (I.e. the identity)
 isCoercion :: Fun -> Bool
 isCoercion = (== "_") -- perhaps this should be changed to "id"?
 
@@ -411,8 +398,8 @@ isNilFun f  = f == "[]"
 isOneFun f  = f == "(:[])"
 isConsFun f = f == "(:)"
 isConcatFun f = f == "(++)"
+
 ------------------------------------------------------------------------------
-type Name = String
 
 -- | Abstract syntax tree.
 newtype Tree = Tree (Fun,[Tree])
@@ -420,34 +407,31 @@ newtype Tree = Tree (Fun,[Tree])
 -- | The abstract syntax of a grammar.
 type Data = (Cat, [(Fun,[Cat])])
 
--- | firstCat returns the first Category appearing in the grammar.
+-- | @firstCat@ returns the first @Cat@egory appearing in the grammar.
 firstCat :: CF -> Cat
-firstCat = valCat . head . rulesOfCF
+firstCat = valCat . head . cfgRules
 
 firstEntry :: CF -> Cat
 firstEntry cf = case allEntryPoints cf of
                  (x:_) -> x
                  _     -> firstCat cf
 
-rulesOfCF   :: CFG f -> [Rul f]
-rulesOfCFP  :: CFP -> [RuleP]
-infoOfCF    :: CFG f -> Info
-pragmasOfCF :: CFG f -> [Pragma]
-
-rulesOfCF   = snd . unCFG
-rulesOfCFP  = rulesOfCF
-infoOfCF    = snd . fst . unCFG
-pragmasOfCF = fst . fst . unCFG
-
 -- aggressively ban nonunique names (AR 31/5/2012)
 
-notUniqueNames :: [Name] -> CF -> [Fun]
-notUniqueNames reserved cf = [head xs | xs <- xss, length xs > 1] where
-  xss = group (sort names)
-  names = reserved ++ allCatsIdNorm cf ++ allFuns cf
-  allFuns g = [ f | f <- map funRule (rulesOfCF g), not (isNilCons f || isCoercion f)]
+-- | Categories and constructors.
+allNames :: CF -> [Fun]
+allNames cf = allCatsIdNorm cf ++
+  [ f | f <- map funRule $ cfgRules cf
+      , not $ isNilCons f
+      , not $ isCoercion f
+  ]
 
--- extract the comment pragmas.
+-- | Get all elements with more than one occurrence.
+filterNonUnique :: (Ord a) => [a] -> [a]
+filterNonUnique xs = [ x | (x:_:_) <- group $ sort xs ]
+
+
+-- | Extract the comment pragmas.
 commentPragmas :: [Pragma] -> [Pragma]
 commentPragmas = filter isComment
  where isComment (CommentS _) = True
@@ -460,34 +444,34 @@ lookupRule f = lookup f . map unRule
 
 -- | Returns all normal rules that constructs the given Cat.
 rulesForCat :: CF -> Cat -> [Rule]
-rulesForCat cf cat = [r | r <- rulesOfCF cf, isParsable r, valCat r == cat]
+rulesForCat cf cat = [r | r <- cfgRules cf, isParsable r, valCat r == cat]
 
 -- | Like rulesForCat but for normalized value categories.
 -- I.e., `rulesForCat (Cat "Exp")` will return rules for category Exp but also
 -- Exp1, Exp2... in case of coercion
 rulesForNormalizedCat :: CF -> Cat -> [Rule]
 rulesForNormalizedCat cf cat =
-    [r | r <- rulesOfCF cf, isParsable r, normCat (valCat r) == cat]
+    [r | r <- cfgRules cf, isParsable r, normCat (valCat r) == cat]
 
 -- | As rulesForCat, but this version doesn't exclude internal rules.
 rulesForCat' :: CF -> Cat -> [Rule]
-rulesForCat' cf cat = [r | r <- rulesOfCF cf, valCat r == cat]
+rulesForCat' cf cat = [r | r <- cfgRules cf, valCat r == cat]
 
 -- | Get all categories of a grammar. (No Cat w/o production returned; No duplicates)
 allCats :: CFG f -> [Cat]
-allCats = nub . map valCat . rulesOfCF
+allCats = nub . map valCat . cfgRules
 
 -- | Gets all normalized identified Categories
 allCatsIdNorm :: CF -> [String]
-allCatsIdNorm = nub . map (identCat . normCat) . allCats
+allCatsIdNorm = nub . map (identCat . normCat . valCat) . cfgRules
 
 -- | Get all normalized Cat
 allCatsNorm :: CF -> [Cat]
-allCatsNorm = nub . map normCat . allCats
+allCatsNorm = nub . map (normCat . valCat) . cfgRules
 
 -- | Is the category is used on an rhs?
 isUsedCat :: CFG f -> Cat -> Bool
-isUsedCat cf cat = cat `elem` [c | r <- rulesOfCF cf, Left c <- rhsRule r]
+isUsedCat cf cat = cat `elem` [c | r <- cfgRules cf, Left c <- rhsRule r]
 
 -- | Group all categories with their rules.
 ruleGroups :: CF -> [(Cat,[Rule])]
@@ -498,36 +482,22 @@ ruleGroupsInternals :: CF -> [(Cat,[Rule])]
 ruleGroupsInternals cf = [(c, rulesForCat' cf c) | c <- allCats cf]
 
 -- | Get all literals of a grammar. (e.g. String, Double)
-literals :: CFG f -> [Cat]
-literals cf = [TokenCat l | l <- lits] ++ owns
- where
-   (lits,_,_,_) = infoOfCF cf
-   owns = map fst (tokenPragmas cf)
-
--- | Get all symbols
-symbols :: CFG f -> [String]
-symbols cf = syms
- where (_,syms,_,_) = infoOfCF cf
+literals :: CFG f -> [TokenCat]
+literals cf = cfgLiterals cf ++ map fst (tokenPragmas cf)
 
 -- | Get the keywords of a grammar.
 reservedWords :: CFG f -> [String]
-reservedWords cf = sort keywords
- where (_,_,keywords,_) = infoOfCF cf
+reservedWords = sort . cfgKeywords
 
 -- | Canonical, numbered list of symbols and reserved words. (These do
 -- not end up in the AST.)
 cfTokens :: CFG f -> [(String,Int)]
-cfTokens cf = zip (sort (symbols cf ++ reservedWords cf)) [1..]
+cfTokens cf = zip (sort (cfgSymbols cf ++ reservedWords cf)) [1..]
 -- NOTE: some backends (incl. Haskell) assume that this list is sorted.
 
--- | Categories that is left-recursive transformable.
-reversibleCats :: CFG f -> [Cat]
-reversibleCats cf = cats
-  where (_,_,_,cats) = infoOfCF cf
-
 -- | Comments can be defined by the 'comment' pragma
 comments :: CF -> ([(String,String)],[String])
-comments cf = case commentPragmas (pragmasOfCF cf) of
+comments cf = case commentPragmas (cfgPragmas cf) of
                xs -> ([p | CommentM p <- xs],
                       [s | CommentS s <- xs])
 
@@ -535,15 +505,15 @@ comments cf = case commentPragmas (pragmasOfCF cf) of
 
 -- | Whether the grammar uses the predefined Ident type.
 hasIdent :: CFG f -> Bool
-hasIdent cf = isUsedCat cf catIdent
+hasIdent cf = isUsedCat cf $ TokenCat catIdent
 
 
 -- these need new datatypes
 
 -- | Categories corresponding to tokens. These end up in the
 -- AST. (unlike tokens returned by 'cfTokens')
-specialCats :: CF -> [Cat]
-specialCats cf = (if hasIdent cf then (TokenCat "Ident":) else id) (map fst (tokenPragmas cf))
+specialCats :: CF -> [TokenCat]
+specialCats cf = (if hasIdent cf then (catIdent:) else id) (map fst (tokenPragmas cf))
 
 -- to print parse trees
 prTree :: Tree -> String
@@ -556,18 +526,18 @@ prTree (Tree (fun,trees)) = fun +++ unwords (map pr2 trees) where
 
 -- * abstract syntax trees: data type definitions
 --
--- The abstract syncax, instanciated by the Data type, is the type signatures
+-- The abstract syntax, instantiated by the Data type, is the type signatures
 -- of all the constructors.
 
 -- | Return the abstract syntax of the grammar.
 -- All categories are normalized, so a rule like:
 --     EAdd . Exp2 ::= Exp2 "+" Exp3 ;
 -- Will give the following signature: EAdd : Exp -> Exp -> Exp
-getAbstractSyntax :: CF -> [(Cat, [(Fun, [Cat])])]
+getAbstractSyntax :: CF -> [Data]
 getAbstractSyntax cf = [ ( c, nub (constructors c) ) | c <- allCatsNorm cf ]
   where
     constructors cat = do
-        rule <- rulesOfCF cf
+        rule <- cfgRules cf
         let f = funRule rule
         guard $ not (isDefinedRule f)
         guard $ not (isCoercion f)
@@ -576,29 +546,29 @@ getAbstractSyntax cf = [ ( c, nub (constructors c) ) | c <- allCatsNorm cf ]
         return (f, cs)
 
 
--- All the function bellow are variation arround the idea of getting the
+-- | All the functions below implement the idea of getting the
 -- abstract syntax of the grammar with some variation but they seem to do a
--- poor job at handling corner cases involving coercions. Use getAbstractSyntax
--- instead if possible.
+-- poor job at handling corner cases involving coercions.
+-- Use 'getAbstractSyntax' instead if possible.
+
 cf2data' :: (Cat -> Bool) -> CF -> [Data]
 cf2data' predicate cf =
-  [(cat, nub (map mkData [r | r <- rulesOfCF cf,
+  [(cat, nub (map mkData [r | r <- cfgRules cf,
                               let f = funRule r,
                               not (isDefinedRule f),
-                              not (isCoercion f), eqCat cat (valCat r)]))
-      | cat <- filter predicate (allCats cf)]
+                              not (isCoercion f), sameCat cat (valCat r)]))
+      | cat <- nub $ map normCat $ filter predicate $ allCats cf ]
  where
-  mkData (Rule f _ its) = (f,[normCat c | Left c <- its, c /= InternalCat])
+  mkData (Rule f _ its) = (f, [normCat c | Left c <- its, c /= InternalCat])
 
 cf2data :: CF -> [Data]
-cf2data = cf2data' isDataCat
-
+cf2data = cf2data' $ isDataCat . normCat
 
 cf2dataLists :: CF -> [Data]
-cf2dataLists = cf2data' isDataOrListCat
+cf2dataLists = cf2data' $ isDataOrListCat . normCat
 
 specialData :: CF -> [Data]
-specialData cf = [(c,[(show c,[TokenCat "String"])]) | c <- specialCats cf] where
+specialData cf = [(TokenCat name, [(name, [TokenCat catString])]) | name <- specialCats cf]
 
 -- to deal with coercions
 
@@ -665,16 +635,8 @@ isEmptyNilRule (Rule f _ ts) = isNilFun f && null ts
 
 -- | Returns the precedence of a category symbol.
 -- E.g.
---
--- >>> precCat (Cat "Abc")
--- 0
---
 -- >>> precCat (CoercCat "Abc" 4)
 -- 4
---
--- But!
--- >>> precCat (ListCat (CoercCat "Abc" 2))
--- 2
 precCat :: Cat -> Integer
 precCat (CoercCat _ i) = i
 precCat (ListCat c) = precCat c
@@ -691,19 +653,19 @@ precCF cf = length (precLevels cf) > 1
 
 
 -- | Does the category have a position stored in AST?
-isPositionCat :: CFG f -> Cat -> Bool
-isPositionCat cf cat =  or [b | TokenReg name b _ <- pragmasOfCF cf, TokenCat name == cat]
+isPositionCat :: CFG f -> TokenCat -> Bool
+isPositionCat cf cat = or [ b | TokenReg name b _ <- cfgPragmas cf, name == cat]
 
 -- | Grammar with permutation profile à la GF. AR 22/9/2004
-type CFP   = CFG FunP -- (Exts,[RuleP])
+type CFP   = CFG FunP
 type FunP  = (Fun,Prof)
-type RuleP = Rul FunP -- (FunP, (Cat, [Either Cat String]))
+type RuleP = Rul FunP
 
 -- | Pair of: the original function name, profile
 type Prof  = (Fun, [([[Int]],[Int])])
 
 cf2cfp :: CF -> CFP
-cf2cfp (CFG (es,rs)) = CFG (es, map cf2cfpRule rs)
+cf2cfp cfg@CFG{..} = cfg { cfgRules = map cf2cfpRule cfgRules }
 
 cf2cfpRule :: Rule -> RuleP
 cf2cfpRule (Rule f c its)  = Rule (f, (f, trivialProf its)) c its
@@ -714,7 +676,7 @@ cfp2cf = fmap fst
 trivialProf :: [Either Cat String] -> [([[Int]],[Int])]
 trivialProf its = [([],[i]) | (i,_) <- zip [0..] [c | Left c <- its]]
 
-{-# DEPRECATED rulesOfCFP, allCatsP, allEntryPointsP  "Use the version without P postfix instead" #-}
+{-# DEPRECATED allCatsP, allEntryPointsP  "Use the version without P postfix instead" #-}
 
 funRuleP :: RuleP -> Fun
 funRuleP = fst . funRule
@@ -723,7 +685,7 @@ ruleGroupsP :: CFP -> [(Cat,[RuleP])]
 ruleGroupsP cf = [(c, rulesForCatP cf c) | c <- allCatsP cf]
 
 rulesForCatP :: CFP -> Cat -> [RuleP]
-rulesForCatP cf cat = [r | r <- rulesOfCFP cf, isParsable r, valCat r == cat]
+rulesForCatP cf cat = [r | r <- cfgRules cf, isParsable r, valCat r == cat]
 
 allCatsP :: CFP -> [Cat]
 allCatsP = allCats
@@ -731,7 +693,7 @@ allCatsP = allCats
 
 -- | Categories that are entry points to the parser
 allEntryPoints :: CFG f -> [Cat]
-allEntryPoints cf = case concat [cats | EntryPoints cats <- pragmasOfCF cf] of
+allEntryPoints cf = case concat [cats | EntryPoints cats <- cfgPragmas cf] of
   [] -> allCats cf
   cs -> cs
 
