@@ -14,32 +14,43 @@
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
-    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+    Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1335, USA
 -}
 
 
-module BNFC.GetCF(parseCF, parseCFP) where
+module BNFC.GetCF where
+
+import Control.Arrow (left)
+import Control.Monad.State (State, evalState, get, modify)
+
+import Data.Char
+import Data.Either  (partitionEithers)
+import Data.Functor ((<$>))
+import Data.List    (nub, partition)
+import Data.Maybe   (mapMaybe)
+
+import qualified Data.Foldable as Fold
+
+import System.Exit (exitFailure)
+import System.IO   (hPutStrLn, stderr)
+
+-- Local imports:
 
 import qualified AbsBNF as Abs
 import ParBNF
+import ErrM
+
 import BNFC.CF
 import BNFC.Options
 import BNFC.TypeChecker
 import BNFC.Utils
 
-import Control.Arrow (left)
-import Control.Monad.State
-import Data.Char
-import Data.Either (partitionEithers)
-import Data.List(nub,partition)
-import Data.Maybe (mapMaybe)
-import ErrM
 
 -- $setup
 -- >>> import PrintBNF
 
 parseCF :: SharedOptions -> Target -> String -> IO CF
-parseCF opts t s = liftM cfp2cf (parseCFP opts t s)
+parseCF opts t s = cfp2cf <$> parseCFP opts t s
 
 parseCFP :: SharedOptions -> Target -> String -> IO CFP
 parseCFP opts target content = do
@@ -49,35 +60,54 @@ parseCFP opts target content = do
                     >>= markTokenCategories
   let cf = cfp2cf cfp
   runErr $ checkDefinitions cf
-  let msgs3 = checkTokens cf
 
-  let reserved = [lang opts | target == TargetJava ]
+  -- Some backends do not allow the grammar name to coincide with
+  -- one of the category or constructor names.
+  let names    = allNames cf
+  () <- when (target == TargetJava && lang opts `elem` names) $
+      die $ unwords $
+        [ "ERROR of backend", show target ++ ":"
+        , "the language name"
+        , lang opts
+        , "conflicts with a name defined in the grammar."
+        ]
 
-  -- Warn of fail if the grammar use non unique names
-  case filter (not . isDefinedRule) $ notUniqueNames reserved cf of
+  -- Warn of fail if the grammar uses non unique names
+  case filter (not . isDefinedRule) $ filterNonUnique names of
     [] -> return ()
-    ns| target `notElem` [TargetHaskell,TargetHaskellGadt,TargetOCaml]
-      -> fail $ "ERROR: names not unique: " ++ unwords ns
-      | otherwise
-      -> do putStrLn $ "Warning: names not unique: " ++ unwords ns
-            putStrLn "This can be an error in other back ends."
+    ns | target `notElem` [TargetHaskell,TargetHaskellGadt,TargetOCaml]
+       -> die $ unlines $
+            [ "ERROR: names not unique: " ++ unwords ns
+            , "This is an error in the backend " ++ show target
+            ]
+       | otherwise
+       -> putStrLn $ unlines $
+            [ "Warning: names not unique: " ++ unwords ns
+            , "This can be an error in other back ends."
+            ]
 
-  -- Print msgs3
-  putStrLn $ unlines msgs3
+  -- Print warnings if user defined nullable tokens.
+  Fold.mapM_ putStrLn $ checkTokens cf
 
   -- Print the number of rules
-  putStrLn $ show (length (rulesOfCF cf)) +++ "rules accepted\n"
+  let nRules = length (cfgRules cf)
+  -- Note: the match against () is necessary for type class instance resolution.
+  () <- when (nRules == 0) $ die $ "ERROR: the grammar contains no rules."
+  putStrLn $ show nRules +++ "rules accepted\n"
 
   -- Print a warning if comment delimiter are bigger than 2 characters
   let c3s = [(b,e) | (b,e) <- fst (comments cf), length b > 2 || length e > 2]
-  unless (null c3s) $do
+  unless (null c3s) $ do
       putStrLn "Warning: comment delimiters longer than 2 characters ignored in Haskell:"
       mapM_ putStrLn [b +++ "-" +++ e | (b,e) <- c3s]
   return cfp
 
   where
     runErr (Ok a) = return a
-    runErr (Bad msg) = fail msg
+    runErr (Bad msg) = die msg
+
+die :: String -> IO a
+die msg = hPutStrLn stderr msg >> exitFailure
 
 {-
     case filter (not . isDefinedRule) $ notUniqueFuns cf of
@@ -115,25 +145,26 @@ getCFP cnf (Abs.Grammar defs0) = do
                   (symbols,keywords) = partition notIdent reservedWords
                   notIdent s         = null s || not (isAlpha (head s)) || any (not . isIdentRest) s
                   isIdentRest c      = isAlphaNum c || c == '_' || c == '\''
-                  reservedWords      = nub [t | r <- rules, Right t <- rhsRule r]
-              in CFG((pragma,(literals,symbols,keywords,[])),rules)
-        revs cf1@(CFG((pragma,(literals,symbols,keywords,_)),rules)) =
-            CFG((pragma,(literals,symbols,keywords,findAllReversibleCats (cfp2cf cf1))),rules)
-    case mapMaybe (checkRule (cfp2cf cf0)) (rulesOfCF cf0) of
+                  reservedWords      = nub [t | r <- rules, isParsable r, Right t <- rhsRule r]
+                     -- Issue #204: exclude keywords from internal rules
+              in CFG pragma literals symbols keywords [] rules
+    case mapMaybe (checkRule (cfp2cf cf0)) (cfgRules cf0) of
       [] -> return ()
       msgs -> fail (unlines msgs)
     return cf0
   where
     (pragma,rules0) = partitionEithers $ concatMap transDef defs
     (defs,inlineDelims) = if cnf then (defs0,id) else removeDelims defs0
+    revs cf1@CFG{..} =
+        cf1 { cfgReversibleCats = findAllReversibleCats (cfp2cf cf1) }
 
 -- | This function goes through each rule of a grammar and replace Cat "X" with
 -- TokenCat "X" when "X" is a token type.
 markTokenCategories :: CFP -> Err CFP
-markTokenCategories (CFG (exts, rules)) = return $ CFG (exts, newRules)
+markTokenCategories cf@CFG{..} = return $ cf { cfgRules = newRules }
   where
-    newRules = [ Rule f (mark c) (map (left mark) rhs) | Rule f c rhs <- rules ]
-    tokenCatNames = [ n | TokenReg n _ _ <- fst exts ] ++ specialCatsP
+    newRules = [ Rule f (mark c) (map (left mark) rhs) | Rule f c rhs <- cfgRules ]
+    tokenCatNames = [ n | TokenReg n _ _ <- cfgPragmas ] ++ specialCatsP
     mark = toTokenCat tokenCatNames
 
 
@@ -159,12 +190,12 @@ removeDelims xs = (ys ++ map delimToSep ds,
     (ds,ys) = partition isDelim xs
     isDelim (Abs.Delimiters{}) = True
     isDelim _ = False
-    
+
     inlineDelim :: Abs.Def -> Either Cat String ->  [Either Cat String]
     inlineDelim (Abs.Delimiters cat open close _ _) (Left c)
       | c == ListCat (transCat cat) = [Right open, Left c, Right close]
     inlineDelim _ x = [x]
-    
+
     inlineDelim' :: Abs.Def -> RuleP -> RuleP
     inlineDelim' d@(Abs.Delimiters cat _ _ _ _) r@(Rule f c rhs)
       | c == ListCat (transCat cat) = r
@@ -176,18 +207,18 @@ removeDelims xs = (ys ++ map delimToSep ds,
     delimToSep (Abs.Delimiters cat _ _ (Abs.SepSepar s) sz) = Abs.Separator  sz cat s
     delimToSep (Abs.Delimiters cat _ _  Abs.SepNone     sz) = Abs.Terminator sz cat ""
     delimToSep x = x
-    
+
 transDef :: Abs.Def -> [Either Pragma RuleP]
 transDef x = case x of
  Abs.Rule label cat items ->
-   [Right $ Rule (transLabel label) (transCat cat) (map transItem items)]
+   [Right $ Rule (transLabel label) (transCat cat) (concatMap transItem items)]
  Abs.Comment str               -> [Left $ CommentS str]
  Abs.Comments str0 str         -> [Left $ CommentM (str0,str)]
  Abs.Token ident reg           -> [Left $ TokenReg (transIdent ident) False reg]
  Abs.PosToken ident reg        -> [Left $ TokenReg (transIdent ident) True reg]
  Abs.Entryp idents             -> [Left $ EntryPoints (map (strToCat .transIdent) idents)]
  Abs.Internal label cat items  ->
-   [Right $ Rule (transLabel label) (transCat cat) (Left InternalCat:map transItem items)]
+   [Right $ Rule (transLabel label) (transCat cat) (Left InternalCat:concatMap transItem items)]
  Abs.Separator size ident str -> map  (Right . cf2cfpRule) $ separatorRules size ident str
  Abs.Terminator size ident str -> map  (Right . cf2cfpRule) $ terminatorRules size ident str
  Abs.Delimiters a b c d e -> map  (Right . cf2cfpRule) $ delimiterRules a b c d e
@@ -275,7 +306,7 @@ coercionRules (Abs.Ident c) n =
 
 ebnfRules :: Abs.Ident -> [Abs.RHS] -> [Rule]
 ebnfRules (Abs.Ident c) rhss =
-  [Rule (mkFun k its) (strToCat c) (map transItem its)
+  [Rule (mkFun k its) (strToCat c) (concatMap transItem its)
      | (k, Abs.RHS its) <- zip [1 :: Int ..] rhss]
  where
    mkFun k i = case i of
@@ -286,10 +317,16 @@ ebnfRules (Abs.Ident c) rhss =
    mkName k s = if all (\c -> isAlphaNum c || elem c ("_'" :: String)) s
                    then s else show k
 
-transItem :: Abs.Item -> Either Cat String
-transItem x = case x of
- Abs.Terminal str   -> Right str
- Abs.NTerminal cat  -> Left (transCat cat)
+-- | Translate a rule item (terminal or non terminal)
+-- It also sanitizes the terminals a bit by skipping empty terminals
+-- or splitting multiwords terminals.
+-- This means that the following rule
+--   Foo. S ::= "foo bar" ""
+-- is equivalent to
+--   Foo. S ::= "foo" "bar"
+transItem :: Abs.Item -> [Either Cat String]
+transItem (Abs.Terminal str) = [Right w | w <- words str]
+transItem (Abs.NTerminal cat) = [Left (transCat cat)]
 
 transCat :: Abs.Cat -> Cat
 transCat x = case x of
@@ -335,16 +372,18 @@ transExp e = case e of
 
 --------------------------------------------------------------------------------
 
---checkTokens :: CFG f -> [String]
-checkTokens cf =
-    if null ns
-    then []
-    else ["Warning : ", -- change to error in a future version
-          "  The following tokens accept the empty string: ",
-          "    "++unwords ns,
-          "  This is error-prone and will not be supported in the future."]
+-- | Check if any of the user-defined terminal categories is nullable.
+checkTokens :: CFG f -> Maybe String
+checkTokens cf
+  | null ns   = Nothing
+  | otherwise = Just $ unlines
+      [ "Warning : "  -- TODO: change to error in a future version
+      , "  The following tokens accept the empty string: "
+      , "    " ++ unwords ns
+      , "  This is error-prone and will not be supported in the future."
+      ]
   where
-    ns = map (show.fst) . filter (nullable.snd) $ tokenPragmas cf
+    ns = map (show . fst) . filter (nullable . snd) $ tokenPragmas cf
 
 -- | Check if a regular expression is nullable (accepts the empty string)
 nullable :: Abs.Reg -> Bool
@@ -382,8 +421,8 @@ checkRule cf (Rule (f,_) cat rhs)
   | badSpecial     = Just $ "Bad special category rule" +++ s
   | badTypeName    = Just $ "Bad type name" +++ unwords (map show badtypes) +++ "in" +++ s
   | badFunName     = Just $ "Bad constructor name" +++ f +++ "in" +++ s
-  | badMissing     = Just $ "No production for" +++ unwords missing ++
-                             ", appearing in rule" +++ s +++ ". Defined categories:" +++ unwords defineds
+  | badMissing     = Just $ "no production for" +++ unwords missing ++
+                             ", appearing in rule\n    " ++ s
   | otherwise      = Nothing
  where
    s  = f ++ "." +++ show cat +++ "::=" +++ unwords (map (either show show) rhs) -- Todo: consider using the show instance of Rule
@@ -401,7 +440,7 @@ checkRule cf (Rule (f,_) cat rhs)
    missing     = filter nodef [show c | Left c <- rhs]
    nodef t = t `notElem` defineds
    defineds =
-    show InternalCat : tokenNames cf ++ specialCatsP ++ map (show . valCat) (rulesOfCF cf)
+    show InternalCat : tokenNames cf ++ specialCatsP ++ map (show . valCat) (cfgRules cf)
    badTypeName = not (null badtypes)
    badtypes = filter isBadType $ cat : [c | Left c <- rhs]
    isBadType (ListCat c) = isBadType c
@@ -427,9 +466,9 @@ checkRule cf (Rule (f,_) cat rhs)
 -- let Ok tree = expandRules (Abs.Grammar [rules1])
 -- in putStrLn (printTree tree)
 -- :}
--- Foo_abc . Foo ::= "abc" ;
--- FooA . Foo ::= A ;
--- Foo1 . Foo ::= "foo" "bar" ;
+-- Foo_abc . Foo ::= "abc";
+-- FooA . Foo ::= A;
+-- Foo1 . Foo ::= "foo" "bar";
 -- Foo2 . Foo ::= "++"
 --
 -- Note that if there are two `rules` macro with the same category, the
@@ -444,7 +483,7 @@ checkRule cf (Rule (f,_) cat rhs)
 -- let Ok tree = expandRules (Abs.Grammar [rules1, rules2])
 -- in putStrLn (printTree tree)
 -- :}
--- Foo1 . Foo ::= "foo" "bar" ;
+-- Foo1 . Foo ::= "foo" "bar";
 -- Foo2 . Foo ::= "foo" "foo"
 --
 -- This is using a State monad to remember the last used index for a category.
@@ -456,7 +495,7 @@ expandRules (Abs.Grammar defs) =
     expand other = return [other]
     mkRule :: Abs.Ident -> Abs.RHS -> State [(String, Int)] Abs.Def
     mkRule ident (Abs.RHS rhs) = do
-      fun <- liftM (Abs.LabNoP . Abs.Id . Abs.Ident) (mkName ident rhs)
+      fun <- Abs.LabNoP . Abs.Id . Abs.Ident <$> mkName ident rhs
       return (Abs.Rule fun (Abs.IdCat ident) rhs)
     mkName :: Abs.Ident -> [Abs.Item] -> State [(String, Int)] String
     mkName (Abs.Ident cat) [Abs.Terminal s]
@@ -464,6 +503,6 @@ expandRules (Abs.Grammar defs) =
     mkName (Abs.Ident cat) [Abs.NTerminal (Abs.IdCat (Abs.Ident s))] =
         return (cat ++ s)
     mkName (Abs.Ident cat) _ = do
-        i <- liftM (maybe 1 (+1) . lookup cat) get
+        i <- maybe 1 (+1) . lookup cat <$> get
         modify ((cat, i):)
         return (cat ++ show i)
