@@ -1,3 +1,4 @@
+{-# LANGUAGE CPP #-}
 -- BNF Converter: Error Monad
 -- Copyright (C) 2004  Author:  Aarne Ranta
 
@@ -7,23 +8,32 @@ module ErrM where
 -- the Error monad: like Maybe type with error msgs
 
 import Control.Monad (MonadPlus(..), liftM)
+#if __GLASGOW_HASKELL__ < 710
 import Control.Applicative (Applicative(..), Alternative(..))
+#else
+import Control.Applicative (Alternative(..))
+#endif
 
 data Err a = Ok a | Bad String
   deriving (Read, Show, Eq, Ord)
 
 instance Monad Err where
   return      = Ok
-  fail        = Bad
   Ok a  >>= f = f a
   Bad s >>= _ = Bad s
+#if __GLASGOW_HASKELL__ < 808
+  fail        = Bad
+#else
+
+instance MonadFail Err where
+  fail = Bad
+#endif
 
 instance Applicative Err where
   pure = Ok
   (Bad s) <*> _ = Bad s
   (Ok f) <*> o  = liftM f o
 
-
 instance Functor Err where
   fmap = liftM
 
